// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// TPlayground implementation.

void TPlayground::cleanup() {
  for(int p=0; p<size(pl); p++) delete pl[p];
  for(int o=0; o<objects.size(); o++) delete objects[o];
  objects.resize(0);
  pl.resize(0);
  }

void TPlayground::generate() {
  int sx = sizeX;
  int sy = sizeY;

  for(int y=0; y<sy; y++) for(int x=0; x<sx; x++)
    g.mapSet(
      x, y, 
      (y == 0 || x == 0 || y == sy-1 || x == sx-1) ? cPermaWall : cFree, 
      0);
  
  TWorldType cw = TWorldType(c.initialWorld);
  
  if(cw == wtEmpty || cw == wtCavern || 
     cw == wtSwissCheese || cw == wtFrame ||
     cw == wtRuins || cw == wtWall)
    for(int u=0; u<c.initBonus; u++) {
      placeBigRandomly(cBonusScore, 0);
      placeBigRandomly(cBonusLevel, 0);
      placeBigRandomly(cBonusAmmo, 0);
      }
  
  if(cw == wtCavern || cw == wtSwissCheese) {
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(g.map[y][x] == cFree) g.mapSet(x, y, cBreakWall, 0);
    }
  
  if(cw == wtCavern) {
    for(int t=0; t<5; t++) {
      int y = 1 + g.rand() % (sy-2);
      int x = 1 + g.rand() % (sx-2);
      if(t < 2) y = sy/2, x = sx/2;
      while(y > 0 && x > 0 && y < sy-1 && x < sx-1) {
        if(g.map[y][x] == cBreakWall) g.mapSet(x, y, cFree, 0);
        if(g.map[y+1][x] == cBreakWall) g.mapSet(x, y+1, cFree, 0);
        if(g.map[y-1][x] == cBreakWall) g.mapSet(x, y-1, cFree, 0);
        if(g.map[y][x+1] == cBreakWall) g.mapSet(x+1, y, cFree, 0);
        if(g.map[y][x-1] == cBreakWall) g.mapSet(x-1, y, cFree, 0);
        int step = g.rand() % 2 * 2 - 1;
        if(g.rand() % 2) x += step; else y += step;
        }        
      }
    }
  
  if(cw == wtConstruction) {
    for(int t=0; t<sy / 5; t++) {
      int y = 1 + g.rand() % (sy-2);
      int x = 1 + g.rand() % (sx-2);
      if(g.map[y][x] != cFree) {t--; continue;}
      int d1 = 0; while(g.map[y-d1][x] == cFree) d1++;
      int d2 = 0; while(g.map[y+d2][x] == cFree) d2++;
      int d3 = 0; while(g.map[y][x-d3] == cFree) d3++;
      int d4 = 0; while(g.map[y][x+d4] == cFree) d4++;

      CellType wall = t & 1 ? cPermaWall : cBreakWall;
      
      for(int d=0; d<d1/2; d++) g.mapSet(x, y-d, wall, 0);
      for(int d=0; d<d2/2; d++) g.mapSet(x, y+d, wall, 0);
      for(int d=0; d<d3/2; d++) g.mapSet(x-d, y, wall, 0);
      for(int d=0; d<d4/2; d++) g.mapSet(x+d, y, wall, 0);
      }
    }
  
  if(cw == wtSwissCheese) {
    for(int t=0; t<sx*sy / 150; t++) {
      int y = 1 + g.rand() % (sy-2);
      int x = 1 + g.rand() % (sx-2);
      for(int ky=y-2; ky<y+3; ky++)
      for(int kx=x-2; kx<x+3; kx++) {
        if(ky<=0 || ky>=sy-1 || kx<=0 || kx>=sx-1) continue;
        if(((ky==y-2)||(ky==y+2)) && ((kx==x-2)||(kx==x+2))) continue;
        if(g.map[ky][kx] == cBreakWall) g.map[ky][kx] = cFree;
        }
      }
    }
  
  if(cw == wtRuins) {
    for(int t=0; t<10; t++) placeBigRandomly(cCannonA, 0);
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(g.map[y][x] == cFree) g.mapSet(x, y, 
        y&x&1 ? cPermaWall :
        g.rand() % 2 == 0 ? cBreakWall : cFree, 0);
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(g.map[y][x] == cCannonA) clearBig(x, y, cFree, 0);
    }
  
  if(cw == wtFrame) {
    int rad = sy / 5;
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(g.map[y][x] == cFree && (y < rad || x < rad || x >= sx-rad || y >= sy-rad))
        g.mapSet(x, y, cBreakWall, 0);
    }
  
  if(cw == wtWall) {
    int r = sx*2/5;
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(g.map[y][x] == cFree && (x >= r && x <= sx-1-r))
        g.mapSet(x, y, cBreakWall, 0);
    }
  
  if(cw == wtGrid) {
    for(int y=1; y<sy-1; y++) for(int x=1; x<sx-1; x++)
      if(y%6 == 0 || x%6 == 0)
      g.mapSet(x, y, cBreakWall, 0);
    }
  
  if(cw == wtGrid || cw == wtConstruction)
    for(int u=0; u<c.initBonus; u++) {
      placeBigRandomly(cBonusScore, 0);
      placeBigRandomly(cBonusLevel, 0);
      placeBigRandomly(cBonusAmmo, 0);
      }
  }

void TPlayground::init(int sx, int sy) {
  cleanup();

  time = 0;
  kqpos = 1;
  seed = systemSeed();
  sizeX = sx;
  sizeY = sy;

  TPlayer *dummy = new TPlayer;
  addPlayer(dummy);
  dummy->init(0xFFFFFF);
  dummy->typeval = -1;
  dummy->typenum = numTypes;

  generate();
  }

int TPlayground::addPlayer(TPlayer* pla) {
  if(size(pl) >= 100) throw TException("too many players!");
  pla->index = size(pl);
  pla->spx = 1 + (rand() % (sizeX - 2));
  pla->spy = 1 + (rand() % (sizeY - 2));
  pl.push_back(pla);
  return pla->index;
  }

void TPlayground::upTime(int newTime) {
  while(time < newTime) {
    makeBackups();
    int timeObj = NEVER;
    int timeKey = NEVER;
    for(int o=0; o<size(objects); o++) {
      int t = objects[o]->getTime();
      if(t < timeObj) timeObj = t;
      }
    if(kqpos != kqend && kqpos % KQSIZE == kqend % KQSIZE)
      throw TException("I have lost my keys");

    if(kqpos != kqend) timeKey = kqueue[kqpos%KQSIZE].time;
    
    time = newTime;
    
    if(timeKey <= time) time = timeKey;
    
    if(timeObj <= time) time = timeObj;
    
    while(kqpos != kqend && time == kqueue[kqpos%KQSIZE].time) {
      TKeypress& h(kqueue[kqpos%KQSIZE]);
      kqpos++;
      if(h.key == keyJoinGame) {
        bool netplayer = h.player & 0x1000000;
        TPlayer *pla = new TPlayer;
        addPlayer(pla);
        pla->init(netplayer ? h.player & 0xFFFFFF: lps[h.player].color);
        if(!netplayer) {
          lps[h.player].index = pla->index;
          }
        }
      else {
        pl[h.player]->key[int(h.key)] = h.pressed;
        for(int o=0; o<size(objects); o++) if(objects[o] && objects[o]->owner == h.player)
          objects[o]->reactToKey(h.key, h.pressed);
        pl[h.player]->reactToKey(h.key, h.pressed);
        }
      }
    
    for(int o=0; o<size(objects); o++) if(objects[o] && objects[o]->getTime() == time)
      objects[o]->go();

    // delete objects...

    int o0 = 0;
    for(int o=0; o<size(objects); o++) if(objects[o]) {
      if(o0 != o) objects[o0] = objects[o];
      o0++;
      }
    
    if(o0 != size(objects)) objects.resize(o0);
    }
  }

void TPlayground::deleteObject(TPlayObject *obj) {
  for(int t=0; t<size(objects); t++) if(objects[t] == obj) {
    delete obj;
    objects[t] = NULL;
    }
  }

#define MULT 305920673

int TPlayground::rand() { 
  long long s0 = seed & ((1<<30) - 1);
  long long s1 = seed >> 30;
  s0 *= MULT;
  s1 *= MULT;
  seed = s0 + (s1 << 30)+ (s1 >> 30);
  seed += seed >> 60;
  seed &= (1LL << 60) - 1;

  return (seed >> 16) & 0x7FFFFFFF;
  }

