// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// TPlayer and TPlayObject implementation.

TPlayObject *TBuffer::getObject() {
  char type;
  get(type);
  switch(type) { 
    case 1:  return new OBullet(*this);
    case 2:  return new OSnakeHead(*this);
    case 3:  return new OThruster(*this);
    case 4:  return new OTetris(*this);
    case 5:  return new OSnakeTail(*this);
    case 6:  return new OCastle(*this);
    case 7:  return new OCannonball(*this);
    case 8:  return new OPad(*this);
    case 9:  return new OBall(*this);
    case 10: return new OBomber(*this);
    case 11: return new OBomb(*this);
    case 12: return new OWorm(*this);
    }
  }

#define numTypes 18

char *typenames[numTypes+1] = {
  "SNAKE",
  "UTRON",
  "UTETRIS",
  "THRUSTER",
  "BOMBER",
  "CASTLE",
  "WORM",
  "PAD",
  "Ball",
  "-break wall-",
  "-perma wall-",
  "-clear-",
  "-cannon-",
  "-bonus ammo-",
  "-bonus level-",
  "-bonus other-",
  "*generate*",
  "*no balls*",
  "!WAIT!"
  };

void TPlayObject::killedBy(int player) {
  if(player != owner) g.pl[player]->frags++;
  ow()->typeval = -1;
  g.deleteObject(this);
  }

void TPlayer::createIcons() {
  for(int n=0; n<numIcons; n++) {
    icons[n].i = d.icons[n].i;
    icons[n].assignColor(color);
    }
  }

void TPlayer::init(int c) {
  color = c;
  lives = 0;
  frags = 0;
  score = 0;
  level = 0;
  ammo  = 10;
  typeval=-1;
  typenum= 0;
  for(int k=0; k<8; k++) key[k] = false;
  createIcons();
  }

void TPlayer::reactToKey(PlayerKey pk, bool pressed) {
  if(!pressed) return;
  if(typeval == -1) {
    if(pk == keyUp    && spy > 1) spy--;
    if(pk == keyDown  && spy < g.sizeY-2) spy++;
    if(pk == keyLeft  && spx > 1) spx--;
    if(pk == keyRight && spx < g.sizeX-2) spx++;
    if(pk == keyEarth) {
      typenum--; if(typenum==-1) typenum = numTypes-1;
      }
    if(pk == keyAir) {
      typenum++; typenum %= numTypes;
      }
    if(pk == keyWater) {
      typenum = 0;
      }
    if(pk == keyFire) {
      // level is reset for each new addition
      switch(typenum) {
        case 0:
          if(g.map[spy][spx] != cFree) return;
          d.playSoundAt(d.sndBirth, spx, spy);
          new OSnakeHead(index, spx, spy, dirAt(spx, spy), true);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;

        case 1:
          if(g.map[spy][spx] != cFree) return;
          d.playSoundAt(d.sndBirth, spx, spy);
          new OSnakeHead(index, spx, spy, dirAt(spx, spy), false);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;

        case 2:
          d.playSoundAt(d.sndBirth, spx, spy);
          new OTetris(index, spx-1, spy);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;

        case 3:
          if(spy < 2 || spy > g.sizeY - 2) return;
          if(spx < 2 || spx > g.sizeX - 2) return;
          d.playSoundAt(d.sndBirth, spx, spy);
          new OThruster(index, spx, spy);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 5000;
          return;
        
        case 4:
          d.playSoundAt(d.sndBirth, spx, spy);
          new OBomber(index, Ctr(spx), Ctr(spy));
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;
        
        case 5:
          if(g.map[spy][spx] != cFree) return;
          g.mapSet(spx, spy, cCastleWall, index);
          d.playSoundAt(d.sndBirth, spx, spy);
          new OCastle(index, spx, spy);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1000;
          return;
          
        case 6:
          d.playSoundAt(d.sndBirth, spx, spy);
          new OWorm(index, spx, spy);
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;

        case 7:
          if(spy < 4 || spy > g.sizeY - 4) return;
          if(spx < 4 || spx > g.sizeX - 4) return;
          d.playSoundAt(d.sndBirth, spx, spy);
          new OPad(index, Ctr(spx), Ctr(spy));
          lives++;
          level = 0;
          ammo = 10;
          typeval = 1;
          return;

        case 8:
          d.playSoundAt(d.sndPlace, spx, spy);
          new OBall(index, Ctr(spx), Ctr(spy), g.rand() % 360);
          return;

        case 9:
          d.playSoundAt(d.sndPlace, spx, spy);
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cBreakWall, index);
          return;

        case 10:
          d.playSoundAt(d.sndPlace, spx, spy);
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cPermaWall, index);
          return;

        case 11:
          d.playSoundAt(d.sndPlace, spx, spy);
          if(mapFreeAt(spx, spy)) g.mapSet(spx, spy, cFree, index);
          return;
        
        case 12:
          d.playSoundAt(d.sndPlace, spx, spy);
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cCannonF, index);
          return;
          
        case 13:
          d.playSoundAt(d.sndRed, spx, spy);
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusAmmo, index);
          return;
          
        case 14:
          d.playSoundAt(d.sndBlue, spx, spy);
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusLevel, index);
          return;
          
        case 15:
          d.playSoundAt(d.sndGreen, spx, spy);
          if(mapFreeBigAt(spx, spy)) placeBig(spx, spy, cBonusScore, index);
          return;
          
        case 16:
          g.generate();
          return;
        
        case 17:
          for(int o=0; o<size(g.objects); o++) if(g.objects[o])
            g.objects[o]->destroyBalls();
          return;
        }
      }
    }
  }

char *TPlayer::getTypeName() { return typenames[typenum]; }

void TPlayer::draw() {
  if(typeval == -1) {
    int dir = (g.time / 4) % 360;
    icons[int(icSpawn)].drawAtRotated(
      d.ofsX + d.Res * spx + d.Res/2, 
      d.ofsY + d.Res * spy + d.Res/2,
      degx[dir], degy[dir]
      );
    }
  }

